/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * main.cpp
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>

#include "main.h"



readanysf::readanysf()

{ 
  fifo = new Fifo( FIFOSIZE  );
  requestcode = R_NOTHING;
  state = STATE_IDLE;
  loop = false;
  eof = true;
  format = -1;
  readsf = NULL;

  AddInAnything();  // add one inlet for any message
  AddOutSignal("audio out Left");
  AddOutSignal("audio out Right");
  AddOutBang();  // add one float outlet (has index 2)

  FLEXT_ADDMETHOD_(0,"start",m_start);
  FLEXT_ADDMETHOD_(0,"play",m_start);
  FLEXT_ADDMETHOD_(0,"pause",m_pause);
  FLEXT_ADDMETHOD_(0,"stop",m_stop);
  FLEXT_ADDMETHOD_(0,"open",m_open);

  FLEXT_ADDMETHOD_(0,"loop",m_loop_s);
  FLEXT_ADDMETHOD_(0,"loop",m_loop_f);

  FLEXT_ADDMETHOD_(0,"pcm_seek",m_pcm_seek);
  FLEXT_ADDMETHOD_(0,"time_seek",m_time_seek);

  FLEXT_ADDBANG(0,m_bang);

  FLEXT_CALLMETHOD(m_child);
} 


readanysf::~readanysf()
{

  mutex.Lock();
  quit = true;
  cond.Signal();
  state = STATE_IDLE;
  if (readsf != NULL)
    delete(readsf);
  if (fifo != NULL)
    delete(fifo);

  mutex.Unlock();

  StopThreads();
}

void readanysf::m_bang() {
  ToOutBang(2);
}

void readanysf::m_loop_s(t_symbol *s) {
  char tmp[7];
  sprintf(tmp,GetString(s));
  if ( strcmp("on", GetString(s)) == 0) {
    post("looping On");
    mutex.Lock();
    loop = true;
    mutex.Unlock();
    return;
  } else if (strcmp("on", GetString(s)) ){
    post("loop Off");
    mutex.Lock();
    loop = false;
    mutex.Unlock();
    return;
  }
}

void readanysf::m_loop_f(float f) {
  if (f == 0.0) {
    post("looping Off");
    mutex.Lock();
    loop = false;
    mutex.Unlock();
    return;
  } else {
    post("looping On");
    mutex.Lock();
    loop = true;
    mutex.Unlock();
    return;
  }
}


void readanysf::m_start() {
  mutex.Lock();
  if (state != STATE_STREAM) {
    if (filename[0] == 0 || readsf == NULL) {
      post("first open a file");
      mutex.Unlock();
      return;
    }
    state = STATE_STREAM;
    eof = false;
    //if ( requestcode != R_PROCESS) {
    requestcode = R_PROCESS;    
    //}
  } else {
    post("already playing");
  }
  mutex.Unlock();
  return;
}

void readanysf::m_time_seek(float f) {
  if (readsf !=NULL) {
    mutex.Lock();
    if ( readsf->TIME_seek( (double)f ) ) {  //  probably dangerous here.  
      fifo->Flush();                         //  better to set here, and then do it 
      requestcode = R_PROCESS;               //  in the FillFifo routine
    }
    mutex.Unlock();
  }
}
void readanysf::m_pcm_seek(int i) {
  if (readsf !=NULL) {
    mutex.Lock();
    if (     readsf->PCM_seek(i) ) {
      fifo->Flush();
      requestcode = R_PROCESS;
    }
    mutex.Unlock();
  }
}
void readanysf::m_stop() {
  if (readsf !=NULL) {
    mutex.Lock();
    fifo->Flush();
    readsf->Rewind();  //  is this dangerous? could do an if(eof) Rewind() at requestcode == R_NOTHING
    requestcode = R_PROCESS;
    state = STATE_IDLE;
    mutex.Unlock();
  }
}

void readanysf::m_pause() {
  if (readsf !=NULL) {
    mutex.Lock();
    state = STATE_IDLE;
    mutex.Unlock();
  }
}

void readanysf::m_open(t_symbol *s) {
  sprintf(filename,GetString(s));
  if (filename[0] == 0) {
    post("bogus filename?");
    return;
  } else {                            //printf("opening: %s\n", filename);
    mutex.Lock();
    state = STATE_IDLE;
    requestcode = R_OPEN;
    mutex.Unlock();
    cond.Signal();
  }
}

void readanysf::FillFifo() {
  int ret, wret;

  //mutex.Lock();   //mutex is locked already
  while ( fifo->FreeSpace() > INOUTSIZE) {  // leave enough space for odd chunks
    mutex.Unlock();
    ret = readsf->Decode();
    mutex.Lock();
    if (ret > 0) {
      wret = fifo->Write( (void *)readsf->buffer, ret*sizeof(float)); //post("wrote %d", wret);
    } else {                         //post("eof, %d", ret);
      if ( readsf->Rewind() ) {      //post("Rewound the file");
	if (!loop) {
	  //state = STATE_IDLE;      //this is premature. we check for eof in m_signal routine
	  eof = true;
	  requestcode = R_NOTHING;
	  //post("not looping");
	  break;
	}
      }
    } 
  }
  //post("buffer full");
}


void readanysf::m_child()
{
  quit = false;

  while(!quit ) {
  //while(!ShouldExit()) {

    mutex.Lock();
    if ( requestcode == R_STOP ) {
      fifo->Flush();
      readsf->Rewind();
      requestcode = R_PROCESS;
      state = STATE_IDLE;
      //post ("stop");
      FillFifo();
      mutex.Unlock();

    } else if ( requestcode == R_OPEN ) {

      if (readsf != NULL) { 
	delete(readsf);
      }
      mutex.Unlock();
      format = SetType(filename);
      mutex.Lock();
      if ( format >= 0) {
	fifo->Flush();
	requestcode = R_PROCESS;
	mutex.Unlock();
	if ( !readsf->Initialize(filename) ) {
	  mutex.Lock();
	  requestcode = R_NOTHING;
	  delete(readsf);
	  readsf = NULL;
	  mutex.Unlock();
	}
	//mutex.Lock();
      } else { // not a recognized file type
	post("file not recognized");
	//mutex.Lock();
	state = STATE_IDLE;
	requestcode = R_NOTHING;
	mutex.Unlock();
      }

    } else if ( requestcode ==  R_PROCESS ) {
      FillFifo(); //take care of mutex locking in FillFifo routine
      mutex.Unlock();
      cond.Wait();
    
    } else { //NOTHING
      //post("do nothing");
      mutex.Unlock();
      cond.Wait();
    }
  }
  

  return;
}




FLEXT_NEW_DSP("readanysf~", readanysf)

void readanysf::m_signal(int n, float *const *in, float *const *out)
{
  float *outs1  = out[0];
  float *outs2  = out[1];
  
  mutex.Lock();      
  if (state == STATE_STREAM) {
    
    //float pdbuf[n*2]; // always multiply by the number of audio signal outlets
    unsigned int size = n * 2 * FLOATSIZE ;
    float * pdbuf = (float *)alloca(size);    

    fifo->Read( (void*)pdbuf, size);
    size = size / FLOATSIZE;
    
    
    if (size < n*2) {
      if (eof) { // end of file, lets close here
	//post("end of file");
	state = STATE_IDLE;
	fifo->Flush();
	requestcode = R_PROCESS;  
	m_bang();  //  bang out at end of file
      } else {
	//post("not EOF, but buffer is empty");
      }
      //if not eof, buffer is stuck somehow so lets keep crunching. 
      //in any case fill from size returned to the end with zeros
      for (int x = size; x < (n*2); x+=2){
	*outs1++  = 0.0;
	*outs2++  = 0.0;
      }
      
    }
    if ( fifo->FreeSpace() > INOUTSIZE ) {
      if ( requestcode == R_PROCESS && !eof) {
	//post("fifo free: %d", fifo->FreeSpace());
	cond.Signal();  // tell the child that the fifo is hungry
      }
    }
    mutex.Unlock();	
    for (int x = 0; x < size; x+=2)  // dont forget to update this with num_channels
      {
	*outs1++  = pdbuf[x];
	*outs2++  = pdbuf[x+1];
      }
    
  } else { // we're not streaming.  just zero out the audio outlets
    mutex.Unlock();	
    while (n--) {
      *outs1++  = 0.0;
      *outs2++  = 0.0;
    }
  }
  
}  


int readanysf::SetType(const char* thefile) {

  char buf[18];
  int fd, fmt;

  fd = open( thefile, O_RDONLY);
  if (fd == -1) {
    fmt = -1;
    readsf = NULL;
    close(fd);
    return -1;
  }

  int bytesread = read(fd, buf, 16);

  if (bytesread < 4) {
    readsf = NULL;
    close(fd);
    return -1;
  }
  if (!strncmp(buf, ".snd", 4)) {
    fmt = FORMAT_NEXT; //, bigendian = 1;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "dns.", 4)) {
    fmt = FORMAT_NEXT; //, bigendian = 0;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "RIFF", 4)) {
    if (bytesread < 12 || strncmp(buf + 8, "WAVE", 4)) {
      readsf = NULL;
      post("bad header ?");
      return -1;
    }
    fmt = FORMAT_WAVE; //, bigendian = 0;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "FORM", 4)) {
    if (bytesread < 12 || strncmp(buf + 8, "AIFF", 4))
      return -2; //goto badheader;
    fmt = FORMAT_AIFF; //, bigendian = 1;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "OggS", 4)) {
    fmt = FORMAT_VORBIS;
    close(fd);
    readsf = new ReadVorbis();
    return fmt;
  } else if (!strncmp(buf, "ID3", 3)) {
    fmt = FORMAT_MAD;
    close(fd);
    readsf = new ReadMad();
    return fmt;
  } else {
    unsigned int sync, bitrate;
    sync     = (unsigned char)buf[0];
    sync     = sync << 3;
    sync     |= ( (unsigned char)buf[1] & 0xE0) >> 5;
    if ( sync == 0x7FF ) {
      fmt = FORMAT_MAD; //printf("mp3 at the beginning\n");
      close(fd);
      readsf = new ReadMad();
      return fmt;
    } else if( !strncasecmp(thefile+strlen(thefile)-4,".mp3",4) ) {
      //trust that its mp3
      fmt = FORMAT_MAD;  //printf("*.mp3 file ending detected....trying mp3\n");
      close(fd);
      readsf = new ReadMad();
      return fmt;
    }
  }
  readsf = NULL;
  return -1;
    
}
